//---------------------------------------------------------------------------
#include <stdio.h>

//#ifdef LINUX
#ifdef LINUX
	#include <stdlib.h>
	#include <unistd.h>
	#include <string.h>
	#include <sys/types.h>
	#include <sys/stat.h>
	#define Sleep(x) sleep(x)
#else
	#include <io.h>
#endif /* LINUX */

#include <fcntl.h>
#include "serial.h"
#include "usb.h"
#include "winusb.h"
#include "commands.h"
#include "errmsgs.h"
#include "ethernet.h"

// writing blocks which fit into the iLCD controllers input queue allow
// to use the Flash routines without connecting the CTS line as well
#define NO_HS_BLOCK_SIZE 	64		// without CTS handshaking
#define HS_BLOCK_SIZE		256	// with handshaking
#define VHS_BLOCK_SIZE		2048	// used for WinUSB and Ethernet only
#define MAX_IN_LEN			128

#define CID	'\xAA'  		// our command introducer

#define SHOW_PROGRESS true

#define FLASH_SIZE_DPC101x			         0x00018000
#define FLASH_SIZE_DPC10xx			         0x00030000
#define USER_DATA_START_DPC20xx	         0x00020000
#define USER_DATA_START_DPC3020	         0xC0000000
#define USER_DATA_START_DPC3050	         0xC0000000
#define USER_DATA_START_DPC3080	         0x90000000
#define USER_DATA_START_DPC3090	         0x90000000
#define USER_DATA_START_DPM5050	      	0x60100000
#define FLASH_SIZE_DPC20xx			         (0x0007D000 - USER_DATA_START_DPC20xx)
#define FLASH_SIZE_DPC3020			         (0x10000 * 16 * 2)	// 2 MByte
#define FLASH_SIZE_DPC3050			         (0x10000 * 16 * 2)	// 2 MByte
#define MAX_PHYS_ADDR				         0xFFFFFFFF
#define NAND_FLASH_BLOCK_SIZE		         (512*32)
#define NAND_FLASH_BLOCK_SIZE_DPC3090		(2048*64)
#define FIRST_NAND_FLASH_PAGE					0x100
#define FLASH_SIZE_DPC3080						(NAND_FLASH_BLOCK_SIZE * (2048 - 3))			// ~ 32 MByte
#define FLASH_SIZE_DPC3090						(NAND_FLASH_BLOCK_SIZE_DPC3090 * (1024 - 3))	// ~ 128 MByte

#define FLASH_BASE_DPM5050						0x60000000
#define FLASH_SIZE_DPM5050						0x1F00000				// 32MB
#define NOR_FLASH_SECTOR_SIZE_DPM5050 		0x1000
#define NOR_FLASH_BLOCK_SIZE_DPM5050		0x10000

#define GENERAL_NAME_LEN		32
#define SHORT_FILE_NAME_LEN	8
#define TIME_LEN				   9
#define DATE_LEN			      9

#define NOTHING_NEW				0x00
#define NEW_PAGE					0x01
#define NEW_SUB_PAGE				0x02

#define swap_word(w)	(word) (((w) / 0x100) + (((w) & 0xFF) << 8))

typedef struct
{
	dword phys_addr;
	dword mem_page;
} mem_mapping_t;

typedef enum
{
	cf_invalid = 0,
	cf_unknown = 0,
	cf_dpc10xx,
	cf_dpc20xx,
	cf_dpc30xx,
	cf_dpm50xx
} chip_family_t;

typedef enum
{
	ci_invalid = 0,
	ci_unknown = 0,
	dpc1010 = 1,
	dpc1030 = 5,
	dpc1050 = 8,
	dpc2060 = 10,
	dpc3020 = 32,
	dpc3050 = 35,
	dpc3080 = 38,
	dpc3090 = 39,
	dpm5050 = 55
} chip_id_t;

typedef enum
{
	ct_invalid = 0,
	ct_graphic,
	ct_font,
	ct_macro,
	ct_appdata,
} contents_t;

// Introduced in iLoader version 3.1:

// *.RID files "embed" raw flash data starting from offset data_start
// with a total size of data_size. The embedded raw flash data has the same
// structure as contained in the *.lcdp-rflash files.

// The program auto-detects if the file in use is a *.RID or a *.lcdp-rflash
// independent of the file extension of the file itself.

// the structure rii_file_header_t is byte-aligned - set packing to 1

#pragma pack(1)


typedef struct
{
	char signature[9];	               // "lcdp-ri?" + 0-char where '?' depends on contents
	dword version;			               // 1
	dword header_size;	               // header size
	byte contents;								// contents description
	byte *data_start;		               // points to data start (relative to beginning of file)
	dword data_size;		               // raw data size
	byte hash[16];			               // hash value (16 byte) of raw data
	char name[GENERAL_NAME_LEN];			// name of element
	char shortname[SHORT_FILE_NAME_LEN];// short file name without extension
	char date[DATE_LEN];					   // creation date
	char time[TIME_LEN];					   // creation time
	char *description;					   // some description
	char *sourcefile;							// full source file name including path
	void *spare;								// points to something not existing yet
} rii_file_header_t;


#pragma pack()

static dword curr_addr = 0;
static dword written = 0;
static dword total_size = 0;
static dword bytes_remaining;
static int hFile;
static bool is_serial_port = true;
static bool is_port_open = false;
static chip_family_t chip_family = cf_dpc10xx;
static chip_id_t chip_id = dpc1050;
static bool no_cts_save;
static dword block_size = HS_BLOCK_SIZE;
static int no_of_backspaces = 0;

commport_t commport = cp_serial;

mem_mapping_t mem_mapping_dpc20xx[] =
{
	{ 0x00000000, 0  },
	{ 0x00001000, 1  },
	{ 0x00002000, 2  },
	{ 0x00003000, 3  },
	{ 0x00004000, 4  },
	{ 0x00005000, 5  },
	{ 0x00006000, 6  },
	{ 0x00007000, 7  },
	{ 0x00008000, 8  },
	{ 0x00010000, 9  },
	{ 0x00018000, 10 },
	{ 0x00020000, 11 },
	{ 0x00028000, 12 },
	{ 0x00030000, 13 },
	{ 0x00038000, 14 },
	{ 0x00040000, 15 },
	{ 0x00048000, 16 },
	{ 0x00050000, 17 },
	{ 0x00058000, 18 },
	{ 0x00060000, 19 },
	{ 0x00068000, 20 },
	{ 0x00070000, 21 },
	{ 0x00078000, 22 },
	{ 0x00079000, 23 },
	{ 0x0007A000, 24 },
	{ 0x0007B000, 25 },
	{ 0x0007C000, 26 },
	{ MAX_PHYS_ADDR, 0  }	// flags end
};

mem_mapping_t mem_mapping_dpc_3020_3050[] =
{
	{ 0xC0000000, 0x80 },
	{ 0xC0010000, 0x81 },
	{ 0xC0020000, 0x82 },
	{ 0xC0030000, 0x83 },
	{ 0xC0040000, 0x84 },
	{ 0xC0050000, 0x85 },
	{ 0xC0060000, 0x86 },
	{ 0xC0070000, 0x87 },
	{ 0xC0080000, 0x88 },
	{ 0xC0090000, 0x89 },
	{ 0xC00A0000, 0x8A },
	{ 0xC00B0000, 0x8B },
	{ 0xC00C0000, 0x8C },
	{ 0xC00D0000, 0x8D },
	{ 0xC00E0000, 0x8E },
	{ 0xC00F0000, 0x8F },
	{ 0xC0100000, 0x90 },
	{ 0xC0110000, 0x91 },
	{ 0xC0120000, 0x92 },
	{ 0xC0130000, 0x93 },
	{ 0xC0140000, 0x94 },
	{ 0xC0150000, 0x95 },
	{ 0xC0160000, 0x96 },
	{ 0xC0170000, 0x97 },
	{ 0xC0180000, 0x98 },
	{ 0xC0190000, 0x99 },
	{ 0xC01A0000, 0x9A },
	{ 0xC01B0000, 0x9B },
	{ 0xC01C0000, 0x9C },
	{ 0xC01D0000, 0x9D },
	{ 0xC01E0000, 0x9E },
	{ 0xC01F0000, 0x9F },
	{ MAX_PHYS_ADDR, 0 }	// flags end
};

//---------------------------------------------------------------------------
static bool Commands_Transmit(byte *data, int len)
{
   switch (commport)
  	{
    	case cp_serial:
      	return Serial_Transmit(data, len);
  		case cp_usb:
			return USB_Transmit(data, len);
		case cp_ethernet:
			return Eth_Transmit(data, len);
#if WIN_USB_SUPPORTED				
		case cp_winusb:
			return WinUSB_Transmit(data, len);
#endif			
		default:       
			return false;
	}
}
//---------------------------------------------------------------------------
static bool Commands_ReceiveChar(byte *c)
{
	switch (commport)
	{
		case cp_serial:
			return Serial_ReceiveChar(c);
		case cp_usb:
			return USB_ReceiveChar(c);
		case cp_ethernet:
			return Eth_ReceiveChar(c);
#if WIN_USB_SUPPORTED				
		case cp_winusb:
			return WinUSB_ReceiveChar(c);
#endif			
		default:      
			return false;
	}
}
//---------------------------------------------------------------------------
static void Commands_SetTimeouts(dword rx_tout, dword tx_tout)
{
	switch (commport)
	{
		case cp_serial:
			Serial_SetTimeouts(rx_tout, tx_tout);
			break;
		case cp_usb:
			USB_SetTimeouts(rx_tout, tx_tout);
			break;
		case cp_ethernet:
			Eth_SetTimeouts(rx_tout, tx_tout);
			break;
#if WIN_USB_SUPPORTED				
		case cp_winusb:
			WinUSB_SetTimeouts(rx_tout, tx_tout);
			break;     
#endif			
	}          
}
//---------------------------------------------------------------------------
// Note: If in_len is > 0 then 'in_len' characters + the final ACK character
// is read into the buffer 'in_data' (must be allocated by the caller) after
// the first ACK (acknowledging the command itself).
// If 'in_len' has a negative value, up to max. 'in_len' characters are read
// after the first ACK (this is used when a non fixed size return string is
// expected).
// If 'in_len' is zero, only the 1st character is read in and checked
// (used for commands without data response).
cmd_result_t Commands_CommonSend(char *cmd, int cmd_len, int *in_len, char *in_data)
{
	byte *data = (byte *) malloc(cmd_len * 2 + 1);
	byte *p = data;
	int i;
	byte c;
	cmd_result_t rv = cr_Ok;
	int expected_len = *in_len;

	if (!is_port_open)
	{
		Error_SetErr(CMD_ERR_PORT_NOT_OPEN);
	}

	// copy command and quote any existing \xAA character
	for (i = 0; i < cmd_len; i++)
	{
		if (cmd[i] == CIN && i > 0)
		{
			// Quote CIN character
			*p++ = CIN;
		}

		*p++ = cmd[i];
	}

	*in_len = 0;

	if (!Commands_Transmit(data, p - data))
	{
		free(data);
		return cr_SendError;
	}

	// not used anymore
	free(data);

	if (!Commands_ReceiveChar(&c))
	{
		if (Error_GetLastErrNo() == SER_ERR_RX_NO_DATA ||
			Error_GetLastErrNo() == USB_ERR_RX_NO_DATA)
		{
			return cr_TimeOut;
		}
		else
			return cr_ReceiveError;
	}

	switch (c)
	{
		case ACK:
			break;
		case NACK:
			return cr_DataError;
		case OVR:
			return cr_Overrun;
		case CSERR:
			return cr_ChkSumError;
		default:
			return cr_Invalid;
	}

	if (expected_len)
	{
		for (i = 0; i < abs(expected_len); i++)
		{
			if (!Commands_ReceiveChar(&c))
			{
				if (Error_GetLastErrNo() == SER_ERR_RX_NO_DATA ||
					Error_GetLastErrNo() == USB_ERR_RX_NO_DATA)
				{
					return cr_TimeOut;
				}
				else
					return cr_ReceiveError;
			}

			if (expected_len < 0 && c == ACK)
			{
				// Terminate with first ACK found
				return cr_Ok;
			}
			else
			{
				// write back received data (even when character received is
				// an ACK)
				in_data[(*in_len)++] = (unsigned char) c;
			}
		}

		if (expected_len > 0)
		{
			// Get the final ACK

			if (!Commands_ReceiveChar(&c))
			{
				if (Error_GetLastErrNo() == SER_ERR_RX_NO_DATA ||
					Error_GetLastErrNo() == USB_ERR_RX_NO_DATA)
				{
					return cr_TimeOut;
				}
				else
					return cr_ReceiveError;
			}

			// Final ACK is not stored into the in_data buffer
			if (c == ACK)
				return cr_Ok;
		}
	}

	return rv;
}
//---------------------------------------------------------------------------
cmd_result_t Commands_SendWTimeout(char *cmd, int cmd_len, dword timeout)
{
	int in_len = 0;
	cmd_result_t rv;

	Commands_SetTimeouts(timeout, DEFAULT_TX_TIMEOUT);
	rv = Commands_CommonSend(cmd, cmd_len, &in_len, NULL);

	// Reset default timeout if neccessary
	if (timeout != DEFAULT_RX_TIMEOUT)
		Commands_SetTimeouts(DEFAULT_RX_TIMEOUT, DEFAULT_TX_TIMEOUT);

	return rv;
}
//---------------------------------------------------------------------------
cmd_result_t Commands_Send(char *cmd, int cmd_len)
{
	return Commands_SendWTimeout(cmd, cmd_len, DEFAULT_RX_TIMEOUT);
}
//---------------------------------------------------------------------------
bool Commands_GetAttention(bool *was_terminal_mode)
{
	char att_term_mode[] = { ESC, '{', '!', 'a' };
	char end_term_mode[] = { ESC, '{', '!', 'e', 'N', 'd' };
	char nop[] = { CID, '\'' };
	int i;

	*was_terminal_mode = false;

	// Send Escape sequence first, the device may be in terminal mode
	*was_terminal_mode =
		Commands_SendWTimeout(att_term_mode, sizeof(att_term_mode), 100) == cr_Ok;

	// try again, the device may not be ready yet after connecting
	if (!*was_terminal_mode)
	{
		*was_terminal_mode =
			Commands_SendWTimeout(att_term_mode, sizeof(att_term_mode), 100) == cr_Ok;
	}

	if (*was_terminal_mode)
	{
		// End terminal mode
		if (Commands_Send(end_term_mode, sizeof(end_term_mode)) == cr_Ok)
			return true;
	}
	else
	{
		// no (valid) response, the device is in standard command mode
		// or not available - test this by sending a nop (twice if necessary)

		for (i = 0; i < 4; i++)
		{
			if (Commands_SendWTimeout(nop, sizeof(nop), 150) == cr_Ok)
				return true;
			
			Sleep(100);
		}
	}

	return false;
}
//---------------------------------------------------------------------------
char *Commands_GetInfo(void)
{
	static char in_data[MAX_IN_LEN + 1] = { 0 };
	char string[] = { CID, '?', 'I' };
	int in_len = MAX_IN_LEN * -1;
	cmd_result_t cr;
	
	Commands_SetTimeouts(DEFAULT_RX_TIMEOUT, DEFAULT_TX_TIMEOUT);
	
	if (Commands_CommonSend(string, sizeof(string), &in_len, in_data) != cr_Ok)
		return NULL;

	return in_data;		
}
//---------------------------------------------------------------------------
char *Commands_GetControllerName(void)
{
	char string[] = { CID, '?', 'C' };
	int in_len = -20;
	static char in_data[20 + 1];
	char *rv = in_data;

	if (Commands_CommonSend(string, sizeof(string), &in_len, in_data) == cr_Ok)
		in_data[in_len] = '\0';
	else
		rv = NULL;

	return rv;
}
//---------------------------------------------------------------------------
// works with DPC20xx and above controllers only
bool Commands_Set_iLCD_BaudRate(dword baud_rate)
{
	char string[] = { CID, 'I', 'B', 0, 0, 0, 0, 0 };

	string[4] = baud_rate >> 24;
	string[5] = baud_rate >> 16;
	string[6] = baud_rate >> 8;
	string[7] = baud_rate;

	return Commands_Send(string, sizeof(string)) == cr_Ok;
}
//---------------------------------------------------------------------------
bool Commands_Reboot(void)
{
	char string[] = { CID, '#' };
	return Commands_Send(string, sizeof(string)) == cr_Ok;
}
//---------------------------------------------------------------------------
static bool Commands_SetFlashPage(char page)
{
	char string[] = { CID, '|', 'P', 0 };

	// last byte contains page
	string[sizeof(string) - 1] = page;

	return Commands_Send(string, sizeof(string)) == cr_Ok;
}
//---------------------------------------------------------------------------
bool Commands_SetNandPage(dword page)
{
	cmd_result_t cr;
	char string[] = { CID, '|', 'p', 0, 0 };

	string[sizeof(string) - 2] = (page >> 8);
	string[sizeof(string) - 1] = page;

	cr = Commands_Send(string, sizeof(string));
	return (cr == cr_Ok);
}
//---------------------------------------------------------------------------
bool Commands_SetNandSubPage(dword mem_sub_page)
{
	cmd_result_t cr;
	char string[] = { CID, '|', 's', 0, 0 };

	string[sizeof(string) - 2] = (mem_sub_page >> 8);
	string[sizeof(string) - 1] = mem_sub_page;

	cr = Commands_Send(string, sizeof(string));
	return (cr == cr_Ok);
}
//---------------------------------------------------------------------------
bool Commands_EraseNandIndex(void)
{
	cmd_result_t cr;
	char string[] = { CID, '|', 'K'};

	// this may take up to 5 seconds in worst case
	cr = Commands_SendWTimeout(string, sizeof(string), 5000);

	return (cr == cr_Ok);
}
//---------------------------------------------------------------------------
bool Commands_EraseFlashPage(char page, word addr)
{
	char string[] = { CID, '|', 'E', 0, 0 };

	if (!Commands_SetFlashPage(page))
		return false;

	// last 2 bytes contain address
	string[sizeof(string) - 2] = addr / 0x100;
	string[sizeof(string) - 1] = addr & 0xFF;

	// this may take up to 5 seconds in worst case
	return Commands_SendWTimeout(string, sizeof(string), 5000) == cr_Ok;
}
//---------------------------------------------------------------------------
bool Commands_EraseNandPage(dword page)
{
	char string[] = { CID, '|', 'E', 0, 0 };
	cmd_result_t cr;
	int i;
	bool rv;

	for (i = 0; i < 3; i++)
	{
		rv = Commands_SetNandPage(page);

		if (rv)
			break;
	}

	if (!rv)
		return false;

	// this may take up to 5 seconds in worst case
	cr = Commands_SendWTimeout(string, sizeof(string), 5000);

	return (cr == cr_Ok);
}
//---------------------------------------------------------------------------
// 16 bit version: this is an NOR-flash alias for Commands_EraseNandPage
bool Commands_Erase16bitFlashPage(dword page)
{
	return Commands_EraseNandPage(page);
}
//---------------------------------------------------------------------------
bool Commands_WriteFlashData(char *data, word size, word addr)
{
	char chksum = 0;
	char intro[] = { CID, '|', 'W', 0, 0, 0, 0 };
	char *full_data = malloc(size + 1 + sizeof(intro));
	int i;
	cmd_result_t cr;

	// Setup address and length
	intro[3] = addr / 0x100;
	intro[4] = addr & 0xff;
	intro[5] = size / 0x100;
	intro[6] = size & 0xff;

	memcpy(full_data, intro, sizeof(intro));
	memcpy(full_data + sizeof(intro), data, size);

	// Calc checksum
	for (i = sizeof(intro) - 4; i < (word) (sizeof(intro) + size); i++)
		chksum += full_data[i];

	*(full_data + sizeof(intro) + size) = ~chksum;

	cr = Commands_SendWTimeout(full_data, sizeof(intro) + size + 1, 500);

	free(full_data);

	return cr == cr_Ok;
}
//---------------------------------------------------------------------------
dword GetMaxFlashWriteSize(void)
{
	byte string[] = { CID, '|', '?', 'W' };
	cmd_result_t cr;
	int in_len = sizeof(word);
	word in_data;

	cr = Commands_CommonSend(string, sizeof(string), &in_len, (char *) &in_data);
	
	if (cr == cr_Ok) 
		return swap_word(in_data);


	if (chip_id >= dpc3080) 
		return 512;

	return 256;
}
//---------------------------------------------------------------------------
// Tell the iLCD controller to write all unwritten data contained in it's
// internal buffer (available/required only for DPC20xx chip family)
bool Commands_FlushAppData(void)
{
	char string[] = { CID, '|', 'L'};

	return Commands_SendWTimeout(string, sizeof(string), 1000) == cr_Ok;
}
//---------------------------------------------------------------------------
bool Commands_VerifyAppData(void)
{
	cmd_result_t cr;
	char string[] = { CID, '|', 'V' };

	cr = Commands_SendWTimeout(string, sizeof(string), 10000);
	return (cr == cr_Ok);
}
//---------------------------------------------------------------------------
// Show return value of the function above in a user readable way
char *Commands_GetResult(cmd_result_t cr)
{
	switch (cr)
	{
		case cr_Ok:
			return "Ok";
		case cr_DataError:
			return "Data error";
		case cr_SendError:
			return "System send error";
		case cr_ReceiveError:
			return "System receive error";
		case cr_Overrun:
			return "Data overrun error";
		case cr_TimeOut:
			return "Timeout error";
		case cr_Invalid:
			return "Invalid response";
		case cr_ChkSumError:
			return "Checksum error";
		default:
			return "Unknown error";
	}
}
//---------------------------------------------------------------------------
static dword Commands_CalcPagedAddr(dword phys_addr, word *short_addr, dword *mem_page, dword *mem_sub_page)
{
	dword block_offset;
	dword new_flags = NOTHING_NEW;
	int i;

	*mem_sub_page = 0;
	
	if (chip_id == dpm5050)
	{
		*mem_page = (phys_addr - FLASH_BASE_DPM5050) / NOR_FLASH_SECTOR_SIZE_DPM5050;
		*short_addr = phys_addr % NOR_FLASH_BLOCK_SIZE_DPM5050;
		new_flags = *short_addr == 0 ? NEW_PAGE : NOTHING_NEW;
	}
	else if (chip_id == dpc3090)
	{
		block_offset = (phys_addr - USER_DATA_START_DPC3090) % NAND_FLASH_BLOCK_SIZE_DPC3090;

#if (NAND_FLASH_BLOCK_SIZE_DPC3090 > 0x10000)
		// we need "sub-pages" here as short_addr can address 64kByte (= 0x10000) only
		*mem_sub_page = block_offset / 0x10000;
		*short_addr = (word) (block_offset % 0x10000);
#else                                                                 
		*short_addr = (phys_addr - USER_DATA_START_DPC3090) % NAND_FLASH_BLOCK_SIZE_DPC3090;
#endif		
			
		*mem_page = ((phys_addr - USER_DATA_START_DPC3090) / NAND_FLASH_BLOCK_SIZE_DPC3090) + FIRST_NAND_FLASH_PAGE;
		new_flags = ((phys_addr - USER_DATA_START_DPC3090) % NAND_FLASH_BLOCK_SIZE_DPC3090) == 0 ? NEW_PAGE : NOTHING_NEW;
		new_flags |= ((phys_addr - USER_DATA_START_DPC3090) % 0x10000) == 0 ? NEW_SUB_PAGE : 0;
	}
	else if (chip_id == dpc3080)
	{
		*short_addr = (phys_addr - USER_DATA_START_DPC3080) % NAND_FLASH_BLOCK_SIZE;
		*mem_page = ((phys_addr - USER_DATA_START_DPC3080) / NAND_FLASH_BLOCK_SIZE) + FIRST_NAND_FLASH_PAGE;
		new_flags = ((phys_addr - USER_DATA_START_DPC3080) % NAND_FLASH_BLOCK_SIZE) == 0 ? NEW_PAGE : NOTHING_NEW;
	}
	else if (chip_id == dpc3020 || chip_id == dpc3050)
	{
		for (i = 0; mem_mapping_dpc_3020_3050[i].phys_addr < MAX_PHYS_ADDR; i++)
		{
			if (phys_addr >= mem_mapping_dpc_3020_3050[i].phys_addr &&
				phys_addr < mem_mapping_dpc_3020_3050[i + 1].phys_addr)
			{
				*mem_page = mem_mapping_dpc_3020_3050[i].mem_page;

				new_flags = (phys_addr == mem_mapping_dpc_3020_3050[i].phys_addr) ? NEW_PAGE : NOTHING_NEW;
				*short_addr = phys_addr - mem_mapping_dpc_3020_3050[i].phys_addr;
			}
		}
	}
	else if (chip_family == cf_dpc20xx)
	{
		for (i = 0; mem_mapping_dpc20xx[i].phys_addr < MAX_PHYS_ADDR; i++)
		{
			if (phys_addr >= mem_mapping_dpc20xx[i].phys_addr &&
				phys_addr < mem_mapping_dpc20xx[i + 1].phys_addr)
			{
				*mem_page = mem_mapping_dpc20xx[i].mem_page;

				new_flags = (phys_addr == mem_mapping_dpc20xx[i].phys_addr) ? NEW_PAGE : NOTHING_NEW;
				*short_addr = phys_addr - mem_mapping_dpc20xx[i].phys_addr;
			}
		}
	}
	else
	{
		*mem_page = (char) ((phys_addr >> 15) + 1);
		*short_addr = (word) ((phys_addr & 0x7FFF) + 0x8000);
		new_flags = (*short_addr == 0x8000) ? NEW_PAGE : NOTHING_NEW;
	}

	return new_flags;
}
//---------------------------------------------------------------------------
static bool Commands_WriteBlock(void)
{
	word short_addr;
	dword mem_page, mem_sub_page;
	char raw_data[max(max(NO_HS_BLOCK_SIZE, HS_BLOCK_SIZE), VHS_BLOCK_SIZE)];
	dword bytes_read_from_file;
	dword bytes_to_write = min(block_size, bytes_remaining);
	long rl;
	dword new_flags;

	new_flags = Commands_CalcPagedAddr(curr_addr, &short_addr, &mem_page, &mem_sub_page);

	if (new_flags)
	{
		if (chip_id == dpm5050 && (new_flags & NEW_PAGE) != 0)
		{
			if (!Commands_Erase16bitFlashPage(mem_page))
			{
				// "Failed to erase Flash memory page"
				return Error_SetErr(CMD_ERR_CANT_ERASE);
			}
		}
		else if (chip_id == dpc3080 || chip_id == dpc3090)
		{
			if (new_flags & NEW_PAGE)
			{
				if (mem_page == FIRST_NAND_FLASH_PAGE)
				{
					if (!Commands_EraseNandIndex())
					{
						// "Failed to erase NAND flash index"
						return Error_SetErr(CMD_ERR_CANT_ERASE_IDX);
					}
				}

				if (!Commands_EraseNandPage(mem_page))
				{
					// "Failed to erase Flash memory page"
					return Error_SetErr(CMD_ERR_CANT_ERASE);
				}
			}
			else if (new_flags & NEW_SUB_PAGE)
			{
				if (!Commands_SetNandSubPage(mem_sub_page))
				{
					// "Failed to set Flash memory sub-page"
					return Error_SetErr(CMD_ERR_CANT_SET_SUB_PAGE);
				}
			}
		}
		else
		{
			if (!Commands_EraseFlashPage((char) mem_page, 0xF000))
			{
				// "Failed to erase Flash memory page"
				return Error_SetErr(CMD_ERR_CANT_ERASE);
			}
		}
	}

	// read raw data from file
	if ((rl = read(hFile, raw_data, bytes_to_write)) < 0)
	{
		// "Failed to read Flash data from file"
		return Error_SetErr(CMD_ERR_CANT_READ_FILE);
	}

	bytes_read_from_file = rl;

	if (bytes_to_write != bytes_read_from_file)
	{
		// "Found unexpected EOF in data file"
		return Error_SetErr(CMD_ERR_REACHED_EOF);
	}

	if (!Commands_WriteFlashData(raw_data, bytes_to_write, short_addr))
	{
		// "Failed to write data to Flash"
		return Error_SetErr(CMD_ERR_CANT_WRITE);
	}

	curr_addr += bytes_to_write;
	written += bytes_to_write;

	bytes_remaining -= bytes_to_write;

#if SHOW_PROGRESS
	if (!(written % 1024) || !bytes_remaining)
	{
		dword percent = written * 100 / total_size;

		while (no_of_backspaces-- > 0)
			putchar('\b');         

		no_of_backspaces = printf("%ldkB (%ld%%) of %ldkB written",
			bytes_remaining ? written / 1024 : total_size / 1024, 
			bytes_remaining ? percent : 100, 
			total_size / 1024);

		fflush(stdout);
	}
#endif

	return true;
}

#ifdef LINUX
//---------------------------------------------------------------------------
int filelength( int fd )
{
	struct stat statbuf;

	if( fstat(fd, &statbuf) < 0 )
		return -1;

	return statbuf.st_size;
}
#endif /* LINUX */

//---------------------------------------------------------------------------
bool Commands_WriteAppData(char *file_name, dword cmd_baud_rate, dword upl_baud_rate)
{
	bool rv = true;
	long fs;
	char header[0x80];
	byte raw_data[max(max(NO_HS_BLOCK_SIZE, HS_BLOCK_SIZE), VHS_BLOCK_SIZE)];
	byte chksum = 0;
	dword bytes_to_check;
	dword i, max_size;
	long rl;
	char *cname;
	bool invalid_controller = false;
	rii_file_header_t rii_file_header;
	bool is_rii_file = true;
	char signature[4];
	dword data_start, data_size;

	if ((hFile = open(file_name, O_RDONLY | O_BINARY)) < 0)
	{
		// "Failed to open Flash data file"
		return Error_SetErr(CMD_ERR_CANT_OPEN_FILE);
	}

	if ((fs = filelength(hFile)) < 0)
	{
		close(hFile);

		// "Failed to get Flash data filesize: "
		return Error_SetErr(CMD_ERR_FILESIZE);
	}

	cname = Commands_GetControllerName();

	block_size = min(GetMaxFlashWriteSize(), block_size);

	// DPC1010 and DPC1011 have only half Flash memory
	if (cname == NULL)
	{
		close(hFile);

		// "Unable to get the name of the iLCD controller"
		return Error_SetErr(CMD_ERR_CANT_GET_CNAME);
	}

	else if (strncmp(cname, "DPM5050", 7) == 0)
	{
		max_size = FLASH_SIZE_DPM5050;
		curr_addr = USER_DATA_START_DPM5050;
		chip_family = cf_dpm50xx;
		chip_id = dpm5050;
	}
	else if (strncmp(cname, "DPC3090", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC3090;
		curr_addr = USER_DATA_START_DPC3090;
		chip_family = cf_dpc30xx;
		chip_id = dpc3090;
	}
	else if (strncmp(cname, "DPC3080", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC3080;
		curr_addr = USER_DATA_START_DPC3080;
		chip_family = cf_dpc30xx;
		chip_id = dpc3080;
	}
	else if (strncmp(cname, "DPC3050", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC3050;
		curr_addr = USER_DATA_START_DPC3050;
		chip_family = cf_dpc30xx;
		chip_id = dpc3050;
	}
	else if (strncmp(cname, "DPC3020", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC3020;
		curr_addr = USER_DATA_START_DPC3020;
		chip_family = cf_dpc30xx;
		chip_id = dpc3020;
	}
	else if (strncmp(cname, "DPC2060", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC20xx;
		curr_addr = USER_DATA_START_DPC20xx;
		chip_family = cf_dpc20xx;
		chip_id = dpc2060;
	}
	else if (strncmp(cname, "DPC1010", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC101x;
		curr_addr = 0;
		chip_family = cf_dpc10xx;
		chip_id = dpc1010;
	}
	else if (strncmp(cname, "DPC1030", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC10xx;
		curr_addr = 0;
		chip_family = cf_dpc10xx;
		chip_id = dpc1030;
	}
	else if (strncmp(cname, "DPC1050", 7) == 0)
	{
		max_size = FLASH_SIZE_DPC10xx;
		curr_addr = 0;
		chip_family = cf_dpc10xx;
		chip_id = dpc1050;
	}
	else
	{
		// Unknown iLCD Controller
		close(hFile);

		// "Unable to get the name of the iLCD controller"
		return Error_SetErr(CMD_ERR_CANT_GET_CNAME);
	}

	if (read(hFile, &rii_file_header, sizeof(rii_file_header_t)) != (int) sizeof(rii_file_header_t))
		is_rii_file = false;

	if (is_rii_file)
	{
		if (strcmp(rii_file_header.signature, "lcdp-rid") != 0 ||
			rii_file_header.version != 1 ||
			rii_file_header.header_size != sizeof(rii_file_header_t) ||
			(contents_t) rii_file_header.contents != ct_appdata)
		{
			is_rii_file = false;
		}

		if (is_rii_file)
		{
			if ((dword) fs == (dword) rii_file_header.data_start + rii_file_header.data_size + 4)
			{
				// RII files end with a "lcdp" signature
				lseek(hFile, (dword) rii_file_header.data_start + rii_file_header.data_size, SEEK_SET);

				if (read(hFile, signature, 4) != 4)
					is_rii_file = false;

				if (is_rii_file && strncmp(signature, "lcdp", 4) != 0)
					is_rii_file = false;
			}
			else
				is_rii_file = false;
		}
	}

	if (is_rii_file)
	{
		// this is the size of the raw data
		data_size = rii_file_header.data_size;

		// this is the position where the raw data starts
		data_start = (dword) rii_file_header.data_start;
		lseek(hFile, data_start, SEEK_SET);
	}
	else
	{
		// File to be checked is not a RII file, check if it is a raw flash file
		// (rewind to start)
		lseek(hFile, 0, SEEK_SET);
		data_start = 0;
		data_size = fs;
	}

	bytes_remaining = data_size;

	if (bytes_remaining > max_size)
	{
		close(hFile);

		// "Data size is too big for the iLCD controller's Flash memory"
		return Error_SetErr(CMD_ERR_FILE_TOO_BIG);
	}

	if (read(hFile, header, sizeof(header)) != (int) sizeof(header))
	{
		close(hFile);

		// "Failed to read Flash data from file"
		return Error_SetErr(CMD_ERR_CANT_READ_FILE);
	}

	// do a minimum of file type checking
	if (memcmp(header, "dplc", 4) != 0 || header[6] != 4)
	{
		close(hFile);

		// "Selected file is not an iLCD Flash data file"
		return Error_SetErr(CMD_ERR_NOT_ILCD);
	}

	if (strncmp(&header[0x6F], "DPM5050", 7) == 0)
	{
		// file is destined to be used with a DPM5050 iLCD controller

		if (chip_id != dpm5050)
			invalid_controller = true;
	}
	else if (strncmp(&header[0x6F], "DPC3090", 7) == 0)
	{
		// file is destined to be used with a DPC3090 iLCD controller

		if (chip_id != dpc3090)
			invalid_controller = true;
	}
	else if (strncmp(&header[0x6F], "DPC3080", 7) == 0)
	{
		// file is destined to be used with a DPC3080 iLCD controller

		if (chip_id != dpc3080)
			invalid_controller = true;
	}
	else if (strncmp(&header[0x6F], "DPC3050", 7) == 0)
	{
		// file is destined to be used with a DPC3050 iLCD controller

		if (chip_id != dpc3050)
			invalid_controller = true;
	}
	else if (strncmp(&header[0x6F], "DPC3020", 7) == 0)
	{
		// file is destined to be used with a DPC3020 iLCD controller

		if (chip_id != dpc3020)
			invalid_controller = true;
	}
	else if (strncmp(&header[0x6F], "DPC20", 5) == 0)
	{
		// file is destined to be used with a DPC20xx iLCD controller

		if (chip_family != cf_dpc20xx)
			invalid_controller = true;
	}
	else if (chip_family != cf_dpc10xx)
		invalid_controller = true;

	if (invalid_controller)
	{
		close(hFile);

		// "Selected file can't be used with the current iLCD controller"
		return Error_SetErr(CMD_ERR_IVAL_CONTR);
	}

	// go to the begin of raw data again
	lseek(hFile, data_start, SEEK_SET);

	// calculate file checksum
	while (bytes_remaining)
	{
		bytes_to_check = min(block_size, bytes_remaining);

		// read raw data from file
		if ((rl = read(hFile, raw_data, bytes_to_check)) < 0)
		{
			close(hFile);

			// "Failed to read Flash data from file"
			return Error_SetErr(CMD_ERR_CANT_READ_FILE);
		}

		if ((long) bytes_to_check != rl)
		{
			close(hFile);

			// "Found unexpected EOF in data file"
			return Error_SetErr(CMD_ERR_REACHED_EOF);
		}

		for (i = 0; i < bytes_to_check; i++)
			chksum += raw_data[i];

		bytes_remaining -= bytes_to_check;
	}

	// Chksum must be zero
	if (chksum)
	{
		close(hFile);

		// "Data file has invalid checksum"
		return Error_SetErr(CMD_ERR_INVAL_CHKSUM);
	}

	bytes_remaining = data_size;

	// go to the begin of raw data again
	lseek(hFile, data_start, SEEK_SET);

	if (upl_baud_rate && upl_baud_rate != cmd_baud_rate)
	{
		if (chip_family == cf_dpc10xx)
		{
			// "Switching upload baud rate is not supported on this chip"
			return Error_SetErr(CMD_ERR_NSUP_SW_BAUD_RATE);
		}

		// try to switch to the upload baudrate first
		if (!Commands_Set_Sys_BaudRate(upl_baud_rate))
		{
			close(hFile);

			// Commands_Set_Sys_BaudRate() prints it's own error message
			return false;
		}

		// obviously the upload baudrate can be used on this PC, switch back to
		// command baud rate first
		if (!Commands_Set_Sys_BaudRate(cmd_baud_rate))
		{
			close(hFile);

			// Commands_Set_Sys_BaudRate() prints it's own error message
			return false;
		}

		// tell the iLCD panel to use the upload baud rate ...
		if (!Commands_Set_iLCD_BaudRate(upl_baud_rate))
		{
			close(hFile);

			// "Failed to switch baud rate"
			return Error_SetErr(CMD_ERR_CANT_SW_BAUD_RATE);
		}

		// ... and switch the PC to the upload baud rate as well
		if (!Commands_Set_Sys_BaudRate(upl_baud_rate))
		{
			close(hFile);

			// "Caution: iLCD device has now upload baudrate"
			return Error_SetErr(CMD_ERR_HAS_UPLD_BAUD_RATE);
		}
	}

	total_size = data_size;
	written = 0;

#if SHOW_PROGRESS
	printf("Programming application data: ");
	
	no_of_backspaces = printf("Total size is %ldkB", total_size / 1024);
	fflush(stdout);
#endif

	while (bytes_remaining)
	{
		if (!Commands_WriteBlock())
		{
			rv = false;
			break;
		}
	}

#if SHOW_PROGRESS
		if (rv)
			printf(" - Ok");
		else
			printf("\nPrograming Error\n");
#endif

	if (rv)
	{
		if (chip_family == cf_dpc20xx || chip_family == cf_dpc30xx || chip_family == cf_dpm50xx)
		{
			// Flush internal programming buffer
			rv = Commands_FlushAppData();

#if SHOW_PROGRESS
			if (rv)
				printf(" - Verifying");
			else
				printf(" - Flushing error\n");
#endif
			// Tell the iLCD device to examine the total checksum
			if (rv)
				rv = Commands_VerifyAppData();

#if SHOW_PROGRESS
			if (rv)
				printf(" - Ok\n");
			else
				printf(" - Verify error\n");
#endif
		}
		else
		{
#if SHOW_PROGRESS
			printf("\n");
#endif
		}
	}

	if (rv && upl_baud_rate && upl_baud_rate != cmd_baud_rate)
	{
		if (!Commands_Set_iLCD_BaudRate(cmd_baud_rate))
		{
			// "Failed to switch baud rate"
			rv = Error_SetErr(CMD_ERR_CANT_SW_BAUD_RATE);
		}

		if (rv && !Commands_Set_Sys_BaudRate(cmd_baud_rate))
		{
			rv = false;
		}
	}

	close(hFile);

	return rv;
}
//---------------------------------------------------------------------------
bool Commands_Set_Sys_BaudRate(dword baud_rate)
{
	if (!is_port_open)
	{
		// "Command port not open"
		return Error_SetErr(CMD_ERR_PORT_NOT_OPEN);
	}

	if (!is_serial_port)
	{
		// "Open command port is not a serial device"
		return Error_SetErr(SER_ERR_NOT_SERIAL);
	}

	return Serial_Config(baud_rate, no_cts_save);
}
//---------------------------------------------------------------------------
bool Commands_PortOpen(char *serial_port, dword baud_rate, bool no_cts, commport_t cp)
{
   commport = cp;

	if (is_port_open)
	{
		// "Command port is already open"
		return Error_SetErr(CMD_ERR_ALREADY_OPEN);
	}

   switch (commport)
  	{
    	case cp_serial:
         // we need this for Commands_Set_Sys_BaudRate()
         no_cts_save = no_cts;

         if (!Serial_Open(serial_port, baud_rate, no_cts))
            return false;

         is_port_open = true;

			block_size = no_cts ? NO_HS_BLOCK_SIZE : HS_BLOCK_SIZE;
			break;

		case cp_usb:
#if WIN_USB_SUPPORTED				
		case cp_winusb:
			if (!WinUSB_Open())		// always try to open WinUSB first
#endif			
			{
				block_size = HS_BLOCK_SIZE;

				if (!USB_Open())
					return false;

				commport = cp_usb;
				is_port_open = true;
				is_serial_port = false;
			}
#if WIN_USB_SUPPORTED				
			else
			{
				// successfully opened WinUSB

				block_size = VHS_BLOCK_SIZE;
				
				commport = cp_winusb;
				is_port_open = true;
				is_serial_port = false;
			}                    
#endif			
			if (!is_port_open)
				return false;
				
			break;
			
		case cp_ethernet:
			block_size = VHS_BLOCK_SIZE;

			if (!Eth_Open())
				return false;

			is_port_open = true;
			is_serial_port = false;
			break;

		default:
      	return false;
	}

	Commands_SetTimeouts(DEFAULT_RX_TIMEOUT, DEFAULT_TX_TIMEOUT);
	return true;
}
//---------------------------------------------------------------------------
bool Commands_PortClose(void)
{
	if (!is_port_open)
	{
		// "Command port is already closed"
		return Error_SetErr(CMD_ERR_ALREADY_CLOSED);
	}

   switch (commport)
  	{
    	case cp_serial:
			return Serial_Close();
  		case cp_usb:
			return USB_Close();
      case cp_ethernet:
			return Eth_Close();
#if WIN_USB_SUPPORTED				
		case cp_winusb:
			return WinUSB_Close();
#endif			
		default:
      	return false;
   }
}
//---------------------------------------------------------------------------

