//---------------------------------------------------------------------------
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/select.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <errno.h>

#include "serial.h"
#include "commands.h"
#include "errmsgs.h"

static int fd_serial = -1;
static int current_baudrate = 0;
static int tx_timeout = 1000;
static int rx_timeout = 1000;


//---------------------------------------------------------------------------
// Set the timeout values of an already opened port
bool Serial_SetTimeouts(dword rx_tout, dword tx_tout)
{
    rx_timeout = rx_tout;
    tx_timeout = tx_tout;
    return true;
}
//---------------------------------------------------------------------------
// Set the serial port values of an already opened port
bool Serial_Config(dword baud_rate, bool no_cts)
{
    int br;
    int r;
    struct termios tios;

    switch (baud_rate)
    {
        case 300:
            br = B300;
            break;
        case 1200:
            br = B1200;
            break;
        case 2400:
            br = B2400;
            break;
        case 4800:
            br = B4800;
            break;
        case 9600:
            br = B9600;
            break;
        case 19200:
            br = B19200;
            break;
        case 38400:
            br = B38400;
            break;
        case 57600:
            br = B57600;
            break;
        default:
            br = B115200;
            break;
    }

    tios.c_cflag = br | CS8 | CLOCAL | CREAD;

    if (!no_cts)
        tios.c_cflag |= CRTSCTS;

    tios.c_iflag = IGNPAR;
    tios.c_lflag &= ~(ECHO | ICANON);
    //tios.c_oflag = OPOST;

    tcflush(fd_serial, TCIFLUSH);
    r = tcsetattr(fd_serial, TCSANOW, &tios);

    current_baudrate = baud_rate;
    return true;
}
//---------------------------------------------------------------------------
// Open serial port - note that the applciation has to wait until data is sent
// and RX data is polled. A better approach would be to use overlapped i/o and
// seperate threads for transmitting/receiving
bool Serial_Open(char *port, dword baud_rate, bool no_cts)
{
    fd_serial = open(port, O_RDWR | O_NOCTTY | O_NDELAY);

    if (fd_serial < 0)
    {
        Error_SetErr(SER_ERR_CANT_OPEN);
        return false;
    }

    Serial_Config(baud_rate, no_cts);
    return true;
}
//---------------------------------------------------------------------------
// Close serial port if it is open
bool Serial_Close()
{
    close(fd_serial);
    fd_serial = -1;
    return true;
}
//---------------------------------------------------------------------------
// Transmit buffer 'data' containing 'size' bytes
// The buffer can include 0-characters as well
bool Serial_Transmit(unsigned char *data, dword size)
{
    int r;
    int done = 0;

    while (done != size)
    {
        fd_set fds;
        struct timeval tv;

        FD_ZERO(&fds);
        FD_SET(fd_serial, &fds);

        tv.tv_sec = 0;
        tv.tv_usec = tx_timeout * 1000;

        r = select(fd_serial + 1, NULL, &fds, NULL, &tv);

        if (r == 1)
        {
            r = write(fd_serial, data + done, size - done);

            if (r <= 0)
            {
                Error_SetErr(SER_ERR_TX);
                return false;
            }
            done += r;
        }
        else
        {
            Error_SetErr(SER_ERR_TX_TIMEOUT);
            return false;
        }
    }
    return true;
}
//---------------------------------------------------------------------------
// Get a byte from the LCD panel - return -1 if no data is available within
// the time setup in Config()
bool Serial_ReceiveChar(unsigned char *c)
{
    int r;
    fd_set fds;
    struct timeval tv;

    FD_ZERO(&fds);
    FD_SET(fd_serial, &fds);

    tv.tv_sec = 0;
    tv.tv_usec = rx_timeout * 1000;

    r = select(fd_serial + 1, &fds, NULL, NULL, &tv);

    if (r == 1)
    {
        r = read(fd_serial, c, 1);
        return true;
    }
    else
    {
        Error_SetErr(SER_ERR_RX);
        return false;
    }
}
//---------------------------------------------------------------------------
bool Serial_IsOpen()
{
    return fd_serial != -1;
}
//---------------------------------------------------------------------------
dword Serial_GetBaudRate()
{
    return current_baudrate;
}
//---------------------------------------------------------------------------
